"""
   Descp: Strategy pattern for proposal's majority outcome

   Created on: 20-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""
from typing import Tuple
import pandas as pd
import numpy as np

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.n_stacked_serie import NStackedSerie
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl


class StProposalMajority(IMetricStrategy):
    __DF_DATE = 'executedAt'
    __DF_PASS = 'winningOutcome'
    __DF_REP = 'totalRepWhenExecuted'
    __DF_VOTES_F = 'votesFor'
    __DF_VOTES_A = 'votesAgainst'
    __DF_BOOST_AT = 'boostedAt'
    __DF_QUORUM = 'queuedVoteRequiredPercentage'
    __DF_INI_COLS = [__DF_DATE, __DF_PASS, __DF_REP, __DF_VOTES_F, __DF_VOTES_A,
                    __DF_BOOST_AT, __DF_QUORUM]

    __DF_MAJORITY_PER = 'majorityPer'
    __DF_IS_ABSOLUTE = 'isAbsolute'
    __DF_COLS = [__DF_DATE, __DF_PASS, __DF_MAJORITY_PER, __DF_IS_ABSOLUTE]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff = dff.dropna(subset=[self.__DF_DATE])
        dff.loc[:, self.__DF_INI_COLS] = dff[self.__DF_INI_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> NStackedSerie:
        if pd_utl.is_an_empty_df(df):
            return NStackedSerie()

        df = self.clean_df(df=df)
        df = self.__calculate_outcome(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        return self.__generate_metric(df=df)


    def __replicate_time_series(self, df: pd.DataFrame, idx: pd.DatetimeIndex) -> pd.DataFrame:
        dff: pd.DataFrame = df

        # joinning all the data in a unique dataframe and fill with NA values
        df3 = pd_utl.get_df_from_lists([idx, None, None, None], self.__DF_COLS)
        df3 = pd_utl.datetime_to_date(df3, self.__DF_DATE)
        # remove duplicated NA
        df3 = pd_utl.drop_duplicate_date_rows(df=dff, dff=df3, date_col=self.__DF_DATE)

        dff = pd.concat([df3, dff], ignore_index=True)
        dff = dff.sort_values(self.__DF_DATE, ignore_index=True)
        dff = dff.fillna(np.nan)

        return dff


    def __generate_metric(self, df: pd.DataFrame) -> NStackedSerie:
        abs_passes, rel_passes = self.__get_sserie_outcome(df=df, has_pass=True)
        abs_fails, rel_fails = self.__get_sserie_outcome(df=df, has_pass=False)
        return NStackedSerie(sseries=[abs_passes, rel_passes, rel_fails, abs_fails])


    def __get_sserie_outcome(self, df: pd.DataFrame, has_pass: bool)\
    -> Tuple[StackedSerie, StackedSerie]:
        if df.empty:
            return (StackedSerie(), StackedSerie())

        first_date = df[self.__DF_DATE].min()
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE, start=first_date)

        # invert has_pass 'cause the df col has True, False and None values.
        dff: pd.DataFrame = pd_utl.filter_by_col_value(df, self.__DF_PASS, 
            (not has_pass), [pd_utl.NEQ])

        # absolute
        d3f: pd.DataFrame = pd_utl.filter_by_col_value(dff, self.__DF_IS_ABSOLUTE, 
            False, [pd_utl.NEQ])
        d3f = self.__replicate_time_series(df=d3f, idx=idx) ## <- HERE THE NONE
        absolute: StackedSerie = self.__get_sserie_from_df(d3f)

        # relative
        d3f: pd.DataFrame = pd_utl.filter_by_col_value(dff, self.__DF_IS_ABSOLUTE, 
            True, [pd_utl.NEQ])
        d3f = self.__replicate_time_series(df=d3f, idx=idx)
        relative: StackedSerie = self.__get_sserie_from_df(d3f)

        return (absolute, relative)


    def __get_sserie_from_df(self, df: pd.DataFrame) -> StackedSerie:
        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        return StackedSerie(serie=serie, y_stack=[df[
            self.__DF_MAJORITY_PER].tolist()])

    def __calculate_outcome(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.astype({
            self.__DF_REP: float,
            self.__DF_VOTES_A: float,
            self.__DF_VOTES_F: float,
            self.__DF_DATE: int,
            self.__DF_QUORUM: int
        })

        # Filter out those with zero
        df = df[df[self.__DF_REP] > 0]

        outcome = df[self.__DF_PASS] == 'Pass'

        # boost == na means there was not boosted
        boost = ~pd.isna(df[self.__DF_BOOST_AT])

        percentage = df[self.__DF_VOTES_F].where(outcome, df[self.__DF_VOTES_A]) / df[self.__DF_REP]
        df[self.__DF_MAJORITY_PER] = round(percentage * 100)

        df[self.__DF_IS_ABSOLUTE] = df[self.__DF_QUORUM] <= df[self.__DF_MAJORITY_PER]

        # winning outcome means more 'votes for' than 'votes against'
        # it passed if outcome and it had relative majority or absolute majority 
        df[self.__DF_PASS] = (df[self.__DF_PASS] == 'Pass') & (boost | df[self.__DF_IS_ABSOLUTE])

        df = df.filter(self.__DF_COLS)
        return df



"""
   Descp: Tester for StProposalOutcome.

   Created on: 14-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List
import pandas as pd

from test.mocks.unix_date_builder import UnixDateBuilder
from dao_analyzer.web.apps.daostack.data_access.daos.metric.strategy.\
    st_proposal_outcome import StProposalOutcome
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.n_stacked_serie import NStackedSerie

class StProposalOutcomeTest(unittest.TestCase):

    def __get_in_data_success_ratio(self) -> pd.DataFrame:
        bl: UnixDateBuilder = UnixDateBuilder()
        boost_date: int = bl.sub(month=3).change(day=2, hour=0, minute=0, second=0).unix()

        return pd.DataFrame([
            #executedAt = today_year-(today_month-3)-25T10:00:00+00:00; boostedAt = today_year-(today_month-3)-02T00:00:00+00:00
            {'executedAt': bl.change(day=25, hour=10).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '25', 'boostedAt': boost_date, 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-(today_month-3)-02T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.change(day=2, hour=0).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '25', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-3)-20T00:00:00+00:00; boostedAt = today_year-(today_month-3)-02T00:00:00+00:00
            {'executedAt': bl.change(day=20).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '0', 'boostedAt': boost_date, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-2)-05T00:00:00+00:00; boostedAt = today_year-(today_month-3)-02T00:00:00+00:00
            {'executedAt': bl.add(month=1).change(day=5).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '1', 'boostedAt': boost_date, 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-(today_month-2)-05T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '1', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-(today_month-1)-28T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.add(month=1).change(day=28).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '30', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-today_month-28T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.add(month=1).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '0', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
        ])


    def test_process_data_boost_outcome(self):
        bl: UnixDateBuilder = UnixDateBuilder()
        bl_b: UnixDateBuilder = UnixDateBuilder()

        bl.sub(month=3).change(day=21, hour=10, minute=0, second=0)
        bl_b.sub(month=3).change(day=2, hour=0, minute=0, second=0)

        in_df: pd.DataFrame = pd.DataFrame([
            #executedAt = today_year-(today_month-3)-21T10:00:00+00:00; boostedAt = today_year-(today_month-3)-02T00:00:00+00:00
            {'executedAt': bl.unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '25', 'boostedAt': bl_b.unix(), 'queuedVoteRequiredPercentage': '50'},

            #executedAt = today_year-(today_month-3)-02T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.change(day=2, hour=0).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '25', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-1)-06T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.add(month=2).change(day=6).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '2', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-(today_month-1)-28T00:00:00+00:00; boostedAt = today_year-(today_month-1)-06T00:00:00+00:00
            {'executedAt': bl.change(day=28).unix(), 'winningOutcome': 'Fail', 'totalRepWhenExecuted': '50', 'votesFor': '2', 'boostedAt': bl_b.add(month=2).change(day=6).unix(), 'queuedVoteRequiredPercentage': '50'},
            
            #executedAt = today_year-today_month-21T00:00:00+00:00; boostedAt = None
            {'executedAt': bl.add(month=1).change(day=21).unix(), 'winningOutcome': 'Pass', 'totalRepWhenExecuted': '50', 'votesFor': '0', 'boostedAt': None, 'queuedVoteRequiredPercentage': '50'},
        ])
        out: List[List[int]] = [
            [1, 0, 0, 0], # fn
            [1, 0, 0, 0], # tp
            [0, 0, 1, 0], # fp
            [0, 0, 1, 1], # tn
        ]

        strategy: StProposalOutcome = StProposalOutcome(m_type=0)
        result: StackedSerie = strategy.process_data(df=in_df)

        for i, l in enumerate(out, 0):
            res_list: List[int] = result.get_i_stack(i_stack=i)
            self.assertListEqual(l, res_list, f'stack = {i}')


    def test_process_data_total_success_ratio(self):
        in_df = self.__get_in_data_success_ratio()
        out: List[int] = [0.3333, 0.5, 0, 1.0]

        strategy: StProposalOutcome = StProposalOutcome(m_type=2)
        result: StackedSerie = strategy.process_data(df=in_df)

        self.assertListEqual(out, result.get_i_stack(0))


    def test_process_data_boost_success_ratio(self):
        in_df = self.__get_in_data_success_ratio()
        out: List[List[int]] = [
            [0.5, 0, None, None],
            [0.0, 1.0, 0, 1.0],
        ]

        strategy: StProposalOutcome = StProposalOutcome(m_type=1)
        result: NStackedSerie = strategy.process_data(df=in_df)

        for i, l in enumerate(out, 0):
            res_list: List[int] = result.get_i_stack(i_value=i, j_stack=0)
            self.assertListEqual(l, res_list, f'stack = {i}')


if __name__ == "__main__":
    unittest.main()



from .imetric_strategy import IMetricStrategy
from .st_assets_tokens import StAssetsTokens
from .st_assets_values import StAssetsValues

__all__ = [
    'IMetricStrategy',
    'StAssetsTokens',
    'StAssetsValues',
]


"""
   Descp: Strategy pattern to create active voters metric.

   Created on: 8-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl


class StActiveVoters(IMetricStrategy):
    __DF_DATE = 'createdAt'
    __DF_VOTER = 'memberAddress'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_VOTER]

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff.loc[:, self.__DF_COLS] = dff[self.__DF_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        # join dates-ids
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        # different voters by month
        df = pd_utl.count_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = df.drop_duplicates(subset=self.__DF_DATE, keep="first")
        df = df.sort_values(self.__DF_DATE)

        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric



from __future__ import print_function as _

import os as _os
import sys as _sys
import json

import dash as _dash

# noinspection PyUnresolvedReferences
from . import _imports_
from ._imports_ import *  # noqa: F401,F403
__all__ = _imports_.__all__

if not hasattr(_dash, '__plotly_dash') and not hasattr(_dash, 'development'):
    print('Dash was not successfully imported. '
          'Make sure you don\'t have a file '
          'named \n"dash.py" in your current directory.', file=_sys.stderr)
    _sys.exit(1)

_basepath = _os.path.dirname(__file__)
_filepath = _os.path.abspath(_os.path.join(_basepath, 'package-info.json'))
with open(_filepath) as f:
    package = json.load(f)

_is_dev = _os.environ.get('DAOA_COMPONENTS_DEV', 'False').lower() == "true"

package_name = package['name'].replace(' ', '_')
json_fname = package_name.replace('-', '_')
python_package_path = __name__.replace('.', '/')
__version__ = package['version']
namespace_name = 'dao_analyzer_components'

_js_rel_file_name = 'dao_analyzer_components' + ('.dev.js' if _is_dev else '.min.js')

_current_path = _os.path.dirname(_os.path.abspath(__file__))

_this_module = _sys.modules[__name__]

async_resources = []

_js_dist = []

_js_dist.extend(
    [
        {
            "relative_package_path": "async-{}.js".format(async_resource),
            "external_url": (
                "https://unpkg.com/{0}@{2}"
                "/{1}/async-{3}.js"
            ).format(package_name, python_package_path, __version__, async_resource),
            "namespace": namespace_name,
            "async": True,
        }
        for async_resource in async_resources
    ]
)

# TODO: Figure out if unpkg link works
_js_dist.extend(
    [
        {
            "relative_package_path": "async-{}.js.map".format(async_resource),
            "external_url": (
                "https://unpkg.com/{0}@{2}"
                "/{1}/async-{3}.js.map"
            ).format(package_name, python_package_path, __version__, async_resource),
            "namespace": namespace_name,
            "dynamic": True,
        }
        for async_resource in async_resources
    ]
)

_js_dist.extend(
    [
        {
            'relative_package_path': _js_rel_file_name,
            'external_url': (
                f'https://unpkg.com/dao-analyzer-components@{__version__}'
                f'/dao_analyzer_components/dao_analyzer_components.min.js'
            ),
            'namespace': namespace_name
        },
        {
            'relative_package_path': _js_rel_file_name + '.map',
            'external_url': (
                f'https://unpkg.com/dao-analyzer-components@{__version__}'
                '/dao_analyzer_components/dao_analyzer_components.min.js.map'
            ),
            'namespace': namespace_name,
        }
    ]
)

_css_dist = [
    {
        'relative_package_path': 'dao_analyzer_components.css',
        'external_url': (
            f'https://unpkg.com/dao-analyzer-components@{__version__}'
            '/dao_analyzer_components/dao_analyzer_components.css'
        ),
        'namespace': namespace_name
    }
]

for _component_name in _imports_.__all__:
    _component = getattr(_imports_, _component_name)
    _component._js_dist = _js_dist
    _component._css_dist = _css_dist



"""
    Descp: Strategy pattern to add the activity of each DAO per month.
           It'll add proposals created and votes and transactions.
    Created on: 26-apr-2022
    Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""

import pandas as pd

from typing import List

from dao_analyzer.web.apps.common.business.transfers import Serie, StackedSerie
from dao_analyzer.web.apps.common.data_access.daos.metric.strategy.imetric_strategy import IMetricStrategy
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl

class StOrganizationActivity(IMetricStrategy):
    __DF_CAST_DATE = 'createdAt'
    __DF_VOTE_DATE = 'startDate'
    __DF_TRANSACTION_DATE = 'date'
    __DF_ID = 'orgAddress'
    __DF_DATE = 'date'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_ID]
    __USED_KEYS = [__DF_ID, __DF_CAST_DATE, __DF_VOTE_DATE, __DF_TRANSACTION_DATE]

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[df.columns.intersection(self.__USED_KEYS)]

    def __get_action(self, df: pd.DataFrame, date_col: str) -> pd.DataFrame:
        if date_col not in df.columns:
            return pd.DataFrame()

        columns: List[str] = [self.__DF_ID, date_col]
        dff = df[columns]
        dff = dff.dropna(subset=[date_col])
        return dff.rename(columns={date_col: self.__DF_DATE})


    def __prepare_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        It combines the different actions in a common action.  
        """
        return pd.concat([
            self.__get_action(df, self.__DF_CAST_DATE),
            self.__get_action(df, self.__DF_VOTE_DATE),
            self.__get_action(df, self.__DF_TRANSACTION_DATE)
        ], ignore_index=True)


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)
        df = self.__prepare_df(df=df)

        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        # join dates-ids
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        # different voters by month
        # This is the only different line from StActiveOrganization
        df = pd_utl.sum_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)
        
        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = df.drop_duplicates(subset=self.__DF_DATE, keep="first")
        df = df.sort_values(self.__DF_DATE)

        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric



"""
   Descp: Strategy pattern to create the percentage of voters from all the users.

   Created on: 05-nov-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import pandas as pd
from typing import List, Tuple

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie 
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_total_members \
    as m_members
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.strategy.st_active_voters \
    as m_voters


class StVotersPercentage(IMetricStrategy):

    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        pass


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()
        
        members_data, vote_data = self.__split_df(df)

        members: StackedSerie = m_members.StTotalMembers()\
            .process_data(df=members_data)

        voters: StackedSerie = m_voters.StActiveVoters()\
            .process_data(df=vote_data)

        percentage: List[float] = self.__calculate_percentage(
            m_members=members,
            m_voters=voters)

        s: Serie = Serie(x=members.get_serie()
            if len(members.get_serie()) >= len(voters.get_serie())
            else voters.get_serie())

        metric: StackedSerie = StackedSerie(
            serie = s, 
            y_stack = [percentage])

        return metric


    def __split_df(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        members_data: pd.DataFrame = df
        vote_data: pd.DataFrame = df.copy() if 'uintVote' in df.columns else pd.DataFrame()

        if 'uintVote' in df.columns:
            # attr of vote df
            vote_data.loc[: ,:] = vote_data[vote_data.uintVote.notna()]
            vote_data = vote_data.dropna(how='all', axis=0)
            members_data.loc[: ,:] = members_data[members_data.uintVote.isna()]
            members_data = members_data.dropna(how='all', axis=0)

        return (members_data, vote_data)


    def __calculate_percentage(self, m_members: StackedSerie, 
    m_voters: StackedSerie) -> List[float]:

        percentage: List[float] = []
        members, voters = self.__adjust_lists(m_members, m_voters)

        for i, _ in enumerate(members):
            numerator: int = voters[i]
            denominator: int = members[i]

            percentage.append(
                round(numerator / denominator * 100, 4) if denominator else None
            )

        return percentage


    def __adjust_lists(self, m_members: StackedSerie, 
    m_voters: StackedSerie) -> Tuple[List[int], List[int]]:

        members: List[int] = m_members.get_i_stack(0)
        voters: List[int] = m_voters.get_i_stack(0)

        s1: List = m_members.get_serie()
        s2: List = m_voters.get_serie()

        fill: List[int] = []

        if len(s1) > len(s2):
            fill = self.__fill_holes(s1=s1, s2=s2)
            voters = fill + voters

        elif len(s1) < len(s2):
            fill = self.__fill_holes(s1=s2, s2=s1)
            members = fill + members

        return (members, voters)


    def __fill_holes(self, s1: List, s2: List) -> List[int]:
        fill: List[int] = []

        for s in s1:
            if len(s2) > 0 and s == s2[0]:
                break
            fill.append(0)

        return fill



"""
   Descp: Dashboard view controller to manage the callbacks.

   Created on: 22-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import Tuple

from dash import html
from dash.dependencies import Input, Output, State
from dao_analyzer.web.apps.common.business.transfers import Organization, OrganizationList, Platform

from dao_analyzer.web.apps.common.presentation.dashboard_view.dashboard_view import _get_dao_info, _get_platform_info, _gen_sum_hdr


# We use organizations Data Access Object to be able to update the organization
# list every callback
def bind_callbacks(app, section_id: str) -> None:
    dao_info_id = section_id + '-info'
    dao_sum_hdr = section_id + '-summary-hdr'

    @app.callback(
        Output(dao_info_id, 'children'),
        Output(dao_sum_hdr, 'children'),
        Input('org-dropdown', 'value'),
        Input('platform-info-store', 'data'),
        State('organization-list-store', 'data'),
    )
    def organization_section_name(value: str, platform_info: dict, org_list: list) -> Tuple[html.Div, html.Div]:
        """This callback is called when the card should change

        Args:
            value (str): The id of the dao selected in the dropdown
            platform_info (dict): The platform info, aggregation of all DAOs
            org_list (list): The list of organizations and its data to be shown on dao-info-container

        Returns:
            Tuple: The dao-info-container and dao card header
        """
        if not value:
            return _get_platform_info(None), _gen_sum_hdr()

        organization_list = OrganizationList.from_json(org_list)        

        if organization_list.is_all_orgs(value):
            return _get_platform_info(Platform.from_json(platform_info)), _gen_sum_hdr()

        result: Organization = next((x for x in organization_list if x.get_id() == value))
        
        return _get_dao_info(result), _gen_sum_hdr(result)

    @app.callback(
        Output(f'{section_id}-body', 'className'),
        Input(f'{section_id}-body', 'className'),
        Input('org-dropdown', 'value'),
        State('organization-list-store', 'data'),
    )
    def show_hide_plots(cname: str, value: str, org_list: list):
        if not value:
            return cname

        if not cname:
            cname = ''

        classes = set(filter(None, cname.split(' ')))
    
        orgs = OrganizationList.from_json(org_list)
        if orgs.is_all_orgs(value):
            # Hide the ones with only-on-dao
            classes.add('is-all-orgs')
        else:
            classes.discard('is-all-orgs')
        
        return ' '.join(classes)



from argparse import ArgumentParser, BooleanOptionalAction, SUPPRESS
from typing import List

from datetime import datetime
import pathlib
import os

from . import config

class CacheScriptsArgParser(ArgumentParser):
    def __init__(self, available_platforms: List[str], available_networks: List[str]):
        super().__init__(description="Main script to populate dao-analyzer cache")

        self.add_argument(
            "-V", "--version",
            action='store_true',
            dest='display_version',
            help="Displays the version and exits"
        )
        self.add_argument(
            "-p", "--platforms",
            choices=available_platforms,
            nargs='*',
            type=str,
            default=available_platforms,
            help="The platforms to update. Every platform is updated by default."
        )
        self.add_argument(
            "--ignore-errors",
            default=True,
            action=BooleanOptionalAction,
            help="Whether to ignore errors and continue")
        self.add_argument(
            "-d", "--debug",
            action='store_true', default=False,
            help="Shows debug info"
        )
        self.add_argument(
            "-f", "--force",
            action='store_true', default=False,
            help="Removes the cache before updating"
        )
        self.add_argument(
            "-F", "--delete-force",
            action="store_true", default=False,
            help="Removes the datawarehouse folder before doing anything"
        )
        self.add_argument(
            "--skip-daohaus-names",
            action="store_true", default=False,
            help="Skips the step of getting Daohaus Moloch's names, which takes some time"
        )
        self.add_argument(
            "--skip-token-balances",
            action="store_true", default=False,
            help="Skips the step of getting every DAO token balances, which takes some time"
        )
        self.add_argument(
            "-n", "--networks",
            nargs="+",
            required=False,
            choices=available_networks,
            default=available_networks,
            help="Networks to update. Every network is updated by default"
        )
        self.add_argument(
            "-c", "--collectors",
            nargs="+",
            required=False,
            help="Collectors to run. For example: aragon/casts"
        )
        self.add_argument(
            "--block-datetime",
            required=False,
            type=datetime.fromisoformat,
            help="Get data up to a block datetime (input in ISO format)"
        )
        self.add_argument(
            "-D", "--datawarehouse",
            help="Specifies the destination folder of the datawarehouse",
            required=False,
            type=pathlib.Path,
            default=config.DEFAULT_DATAWAREHOUSE
        )
        self.add_argument(
            "--cc-api-key",
            help="Set the CryptoCompare API key (overrides environment variable)",
            required=False,
            type=str,
            default=os.getenv('DAOA_CC_API_KEY')
        )
        self.add_argument(
            "--only-updatable",
            help=SUPPRESS, # "Run only updatable collectors (only for testing)",
            action='store_true',
            required=False,
            default=False
        )
        self.add_argument(
            "--daostack-all",
            help="Obtain all DAOs in DAOstack, not only registered ones",
            action='store_true',
            required=False,
            default=False,
        )


from setuptools import setup, find_packages, find_namespace_packages
from functools import partial

def version_dev(version):
    from setuptools_scm.version import get_local_node_and_date

    if version.dirty:
        return get_local_node_and_date(version)
    else:
        return ""

package_dir = {
  'dao_analyzer.web': 'dao_analyzer/web',
  'dao_analyzer.cache_scripts':  'cache_scripts',
  # Unfortunately, this package can't be in the same namespace as the others
  # see https://github.com/plotly/dash/issues/2236
  'dao_analyzer_components': 'dao_analyzer_components/dao_analyzer_components',
}

def util_add_prefix_to_list(prefix, package_list):
    return [ '.'.join([prefix, p]) for p in package_list]

def custom_f_packages(f, *args, **kwargs):
    package_list = []
    for p, d in package_dir.items():
        package_list.append(p)
        package_list.extend(util_add_prefix_to_list(p, f(d)))
        
    return package_list

custom_find_packages = partial(custom_f_packages, find_packages)
custom_find_namespace_packages = partial(custom_f_packages, find_namespace_packages)

def main():
    setup(
        # Fix in case we need to build sdist instead
        use_scm_version={
            'local_scheme': version_dev,
            'write_to': 'dao_analyzer/web/_version.py',
        },
        setup_requires=['setuptools_scm'],
        packages = custom_find_packages(),
        namespace_packages = ['dao_analyzer'],
        package_dir = package_dir,
    )

if __name__ == "__main__":
    main()



"""
   Descp: This file stores the common path of the datawarehouse

   Created on: 28-sep-2022

   Copyright 2022 David Davó
        <david@ddavo.me>
"""
from pathlib import Path
import os

DATAWAREHOUSE: Path = Path(os.getenv('DAOA_DW_PATH', './datawarehouse'))



"""
   Descp: ApiRequester mock up for testing

   Created on: 14-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import Any, List, Dict
from dao_analyzer.web.apps.common.data_access.requesters.irequester import IRequester

class ApiRequesterMock(IRequester):
    def __init__(self, any_list: List[Dict]):
        self.call_times: int = 0
        self.any_list: List[Dict] = any_list


    def request(self, any: Any) -> Dict:
        result = self.any_list[self.call_times]

        if self.call_times < len(self.any_list)-1:
            self.call_times += 1

        return result


    def get_elems_per_chunk(self, n_chunk: int) -> int:
        return pow(2, n_chunk)



"""
   Descp: This class is used to represent a bar plot.

   Created on: 09-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import Dict
import plotly.graph_objs as go

from dao_analyzer.web.apps.common.presentation.charts.layout.figure.figure import Figure
import dao_analyzer.web.apps.common.resources.colors as Color

class BarFigure(Figure):

    def __init__(self) -> None:
        super().__init__()
        super().configuration.add_x_params(params={'tickangle': True})
        super().configuration.add_y_params(params={'grid': True})


    @staticmethod
    def get_empty_plot_data() -> Dict:
        return {
            'x': [],
            'y': [],
            'name': '',
            'color': Color.LIGHT_BLUE,
            'type': '-',
            'x_format': '',
        }


    def get_empty_figure(self) -> Dict:
        return self.get_figure(plot_data = self.get_empty_plot_data())


    def get_figure(self, plot_data: Dict) -> Dict:
        """
        Returns the bar chart filled with plot_data.
        Arguments:
            * plot_data = {
                'x': a list type filled with values.
                'y': a list type filled with values.
                'name': the name of the bar chart.
                'color': The bar's color.
                'type': The type of the x-axis values, e.g. date.
                'x_format': Format of the x-axis values, e.g. '%b, %Y'
            } 
        """
        bar: go.Bar = go.Bar(
                x=plot_data['x'],
                y=plot_data['y'],
                name=plot_data['name'],
                marker_color=plot_data['color'],
        )

        super().configuration.add_x_params(
            params={
                # 'tickvals': plot_data['x'],
                'type': plot_data['type'],
                'tickformat': plot_data['x_format'],
            })

        super().configuration.disable_legend()

        layout: go.Layout = go.Layout(
            xaxis=super().configuration.get_x_axis_layout(),
            yaxis=super().configuration.get_y_axis_layout(),
            legend=super().configuration.get_legend(),
            shapes=super().configuration.get_shapes(),
            margin=self.configuration.get_margin(),
        )

        return {'data': [bar], 'layout': layout}



"""
   Descp: This class is used to build a unix date.

   Created on: 7-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from datetime import datetime
from dateutil.relativedelta import relativedelta as rd

class UnixDateBuilder():
    UNIX_SPACER: datetime = datetime(1970, 1, 1)

    def __init__(self):
        self._date = datetime.now()


    def set_date(self, date: datetime) -> None:
        self._date = date


    def add(self, year: int = 0, month: int = 0, day: int = 0, 
        hour: int = 0, minute: int = 0, second: int = 0):
        
        if year > 0:
            self._date = self._date + rd(years=+year)
        if month > 0:
            self._date = self._date + rd(months=+month)
        if day > 0:
            self._date = self._date + rd(days=+day)
        if hour > 0:
            self._date = self._date + rd(hours=+hour)
        if minute > 0:
            self._date = self._date + rd(minutes=+minute)
        if second > 0:
            self._date = self._date + rd(seconds=+second)
        
        return self


    def sub(self, year: int = 0, month: int = 0, day: int = 0, 
        hour: int = 0, minute: int = 0, second: int = 0) -> None:
        
        if year > 0:
            self._date = self._date + rd(years=-year)
        if month > 0:
            self._date = self._date + rd(months=-month)
        if day > 0:
            self._date = self._date + rd(days=-day)
        if hour > 0:
            self._date = self._date + rd(hours=-hour)
        if minute > 0:
            self._date = self._date + rd(minutes=-minute)
        if second > 0:
            self._date = self._date + rd(seconds=-second)

        return self


    def change(self, year: int = 0, month: int = 0, day: int = 0, 
        hour: int = -1, minute: int = -1, second: int = -1) -> None:
        
        if year > 0:
            self._date = self._date + rd(year=year)
        if month > 0:
            self._date = self._date + rd(month=month)
        if day > 0:
            self._date = self._date + rd(day=day)
        if hour >= 0:
            self._date = self._date + rd(hour=hour)
        if minute >= 0:
            self._date = self._date + rd(minute=minute)
        if second >= 0:
            self._date = self._date + rd(second=second)

        return self


    def unix(self) -> int:
        return (self._date - self.UNIX_SPACER).total_seconds()



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DataPoint(Component):
    """A DataPoint component.
Used from each DataPoint on the right side of the second header

Keyword arguments:

- id (string; optional):
    The css id to use.

- evolution (number; optional):
    The absolute difference between this month and the last.

- evolution_rel (number; optional):
    The relative difference between this month and the last.

- number (number | string; default NaN):
    The big number to show.

- title (string; required):
    The title of the datapoint."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dao_analyzer_components'
    _type = 'DataPoint'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, title=Component.REQUIRED, number=Component.UNDEFINED, evolution=Component.UNDEFINED, evolution_rel=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'evolution', 'evolution_rel', 'number', 'title']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'evolution', 'evolution_rel', 'number', 'title']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        for k in ['title']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')

        super(DataPoint, self).__init__(**args)



"""
    Descp: Aragon Runner and Collectors

    Created on: 02-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
from typing import List

import pkgutil
from gql.dsl import DSLField
import pandas as pd
import numpy as np
import json

from ..aragon import __name__ as aragon_module_name
from ..common.cryptocompare import CCPricesCollector
from ..common import ENDPOINTS, Collector
from ..common.graphql import GraphQLCollector, GraphQLRunner
from ..common.blockscout import BlockscoutBallancesCollector

class AppsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('apps', runner, endpoint=ENDPOINTS[network]['aragon'], network=network)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.apps(**kwargs).select(
            ds.App.id,
            ds.App.isForwarder,
            ds.App.isUpgradeable,
            ds.App.repoName,
            ds.App.repoAddress,
            ds.App.organization.select(ds.Organization.id)
        )

class BalancesCollector(BlockscoutBallancesCollector):
    def __init__(self, runner, base, network: str):
        super().__init__(runner, addr_key='recoveryVault', base=base, network=network)

class CastsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('casts', runner, endpoint=ENDPOINTS[network]['aragon_voting'], network=network, pbar_enabled=False)

        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            df = df.rename(columns={
                'voterId':'voter', 
                'voteAppAddress':'appAddress',
                'voteOrgAddress':'orgAddress'})
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.casts(**kwargs).select(
            ds.Cast.id,
            ds.Cast.vote.select(ds.Vote.id),
            ds.Cast.voter.select(ds.Voter.id),
            ds.Cast.supports,
            ds.Cast.stake,
            ds.Cast.createdAt,
            ds.Cast.vote.select(
                ds.Vote.orgAddress,
                ds.Vote.appAddress
            )
        )

class OrganizationsCollector(GraphQLCollector):
    DAO_NAMES=pkgutil.get_data(aragon_module_name, 'dao_names.json')

    def __init__(self, runner, network: str):
        super().__init__('organizations', runner, endpoint=ENDPOINTS[network]['aragon'], network=network)

        @self.postprocessor
        def set_dead_recoveryVault(df: pd.DataFrame) -> pd.DataFrame:
            df['recoveryVault'] = df['recoveryVault'].replace(r'^0x0+$', np.NaN, regex=True)
            return df

        @self.postprocessor
        def apply_names(df: pd.DataFrame) -> pd.DataFrame:
            names_dict = json.loads(self.DAO_NAMES)

            if self.network not in names_dict.keys() or \
            not names_dict[self.network] or \
            df.empty:
                return df

            names_df = pd.json_normalize(names_dict[self.network])
            names_df['id'] = names_df['address'].str.lower()
            names_df['name'] = names_df['name'].fillna(names_df['domain'])
            names_df = names_df[['id', 'name']]
            df = df.merge(names_df, on='id', how='left')

            return df

        @self.postprocessor
        def copy_id(df: pd.DataFrame) -> pd.DataFrame:
            df['orgAddress'] = df['id']
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.organizations(**kwargs).select(
            ds.Organization.id,
            ds.Organization.createdAt,
            ds.Organization.recoveryVault
        )

class MiniMeTokensCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('miniMeTokens', runner, endpoint=ENDPOINTS[network]['aragon_tokens'], network=network, pbar_enabled=False)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.miniMeTokens(**kwargs).select(
            ds.MiniMeToken.id,
            ds.MiniMeToken.address,
            ds.MiniMeToken.totalSupply,
            ds.MiniMeToken.transferable,
            ds.MiniMeToken.name,
            ds.MiniMeToken.symbol,
            ds.MiniMeToken.orgAddress,
            ds.MiniMeToken.appAddress,
            ds.MiniMeToken.lastUpdateAt
        )

class TokenHoldersCollector(GraphQLCollector):
    def __init__(self, runner: GraphQLRunner, network: str):
        super().__init__('tokenHolders', runner, endpoint=ENDPOINTS[network]['aragon_tokens'], network=network)

        @self.postprocessor
        def add_minitokens(df: pd.DataFrame) -> pd.DataFrame:
            tokens = runner.filterCollector(name='miniMeTokens', network=network).df
            tokens = tokens.rename(columns={'address':'tokenAddress', 'orgAddress':'organizationAddress'})
            return df.merge(tokens[['tokenAddress', 'organizationAddress']], on='tokenAddress', how='left')
            
    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.tokenHolders(**kwargs).select(
            ds.TokenHolder.id,
            ds.TokenHolder.address,
            ds.TokenHolder.tokenAddress,
            ds.TokenHolder.lastUpdateAt,
            ds.TokenHolder.balance
        )

class TokenPricesCollector(CCPricesCollector):
    pass

class ReposCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('repos', runner, network=network, endpoint=ENDPOINTS[network]['aragon'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.repos(**kwargs).select(
            ds.Repo.id,
            ds.Repo.address,
            ds.Repo.name,
            ds.Repo.node,
            ds.Repo.appCount
        )

class TransactionsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('transactions', runner, network=network, endpoint=ENDPOINTS[network]['aragon_finance'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.transactions(**kwargs).select(
            ds.Transaction.id,
            ds.Transaction.orgAddress,
            ds.Transaction.appAddress,
            ds.Transaction.token,
            ds.Transaction.entity,
            ds.Transaction.isIncoming,
            ds.Transaction.amount,
            ds.Transaction.date,
            ds.Transaction.reference
        )

class VotesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('votes', runner, network=network, endpoint=ENDPOINTS[network]['aragon_voting'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.votes(**kwargs).select(
            ds.Vote.id,
            ds.Vote.orgAddress,
            ds.Vote.appAddress,
            ds.Vote.creator,
            ds.Vote.originalCreator,
            ds.Vote.metadata,
            ds.Vote.executed,
            ds.Vote.executedAt,
            ds.Vote.startDate,
            ds.Vote.supportRequiredPct,
            ds.Vote.minAcceptQuorum,
            ds.Vote.yea,
            ds.Vote.nay,
            ds.Vote.voteNum,
            ds.Vote.votingPower
        )

class AragonRunner(GraphQLRunner):
    name: str = 'aragon'

    def __init__(self, dw=None):
        super().__init__(dw)
        self._collectors: List[Collector] = []
        ## TODO: Fix aragon-tokens xdai subgraph and redeploy
        self.networks = ['mainnet']

        for n in self.networks: 
            self._collectors.extend([
                AppsCollector(self, n),
                CastsCollector(self, n),
                MiniMeTokensCollector(self, n),
                ReposCollector(self, n),
                TransactionsCollector(self, n),
                TokenHoldersCollector(self, n),
                VotesCollector(self, n)
            ])
            oc = OrganizationsCollector(self, n)
            bc = BalancesCollector(self, oc, n)
            self._collectors += [oc, bc]
        
        self._collectors.append(CCPricesCollector(self))

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors


"""
   Descp: Strategy pattern to calculate active users, the ones who performs
   at least one action (create a proposal, vote or stake) in a given month. 

   Created on: 24-aug-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List
import pandas as pd

from dao_analyzer.web.apps.common.data_access.daos.metric.strategy import IMetricStrategy

from dao_analyzer.web.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.web.apps.common.business.transfers.serie import Serie
import dao_analyzer.web.apps.common.data_access.pandas_utils as pd_utl


class StActiveUsers(IMetricStrategy):
    __DF_DATE = 'createdAt'
    __DF_PROPOSER = 'proposer'
    __DF_VOTER = 'voter'
    __DF_STAKER = 'staker'
    __DF_USER = 'user'
    __DF_COUNT = 'activeUsers'
    __DF_COLS = [__DF_DATE, __DF_USER]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        actioners = set([self.__DF_PROPOSER, self.__DF_VOTER, self.__DF_STAKER])
        actioners = list(actioners.intersection(set(df.columns)))
        used_cols: List[str] = actioners + [self.__DF_DATE]
        dff = dff[used_cols]
        return dff


    def __get_user_action(self, df: pd.DataFrame, actioner: str) -> pd.DataFrame:
        if actioner not in df.columns:
            return pd.DataFrame()

        columns: List[str] = [self.__DF_DATE, actioner]
        dff = df[columns]
        dff = dff.dropna(subset=[actioner])
        return dff.rename(columns={actioner: self.__DF_USER})


    def __prepare_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Changes the columns proposer, voter and staker into an unique column
        named user.
        """
        return pd.concat([
            self.__get_user_action(df, self.__DF_PROPOSER),
            self.__get_user_action(df, self.__DF_VOTER),
            self.__get_user_action(df, self.__DF_STAKER)
        ], ignore_index=True)


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)
        df = self.__prepare_df(df=df)

        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        # join dates-ids
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        # different users by month
        df = pd_utl.count_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)

        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = df.drop_duplicates(subset=self.__DF_DATE, keep="first")
        df = df.sort_values(self.__DF_DATE)

        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric



"""
   Descp: Manage the application logic, and it's used to interconect the
        data_access and presentation layers.

   Created on: 2-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List, Callable
from dash import html
import dao_analyzer_components as dac

from dao_analyzer.web.apps.common.business.transfers.organization import Platform, OrganizationList

from dao_analyzer.web.apps.common.presentation.charts.dt_controller import DataTableController
import dao_analyzer.web.apps.common.presentation.dashboard_view.dashboard_view as view
import dao_analyzer.web.apps.common.presentation.dashboard_view.controller as view_cont
from dao_analyzer.web.apps.common.data_access.daos.platform_dao\
    import PlatformDAO
from dao_analyzer.web.apps.daohaus.business.metric_adapter.asset_values import AssetsValues
from dao_analyzer.web.apps.daohaus.business.metric_adapter.asset_tokens import AssetsTokens
from dao_analyzer.web.apps.common.presentation.charts.chart_controller import ChartController
from dao_analyzer.web.apps.common.presentation.charts.chart_sum_controller import ChartSummaryController
from dao_analyzer.web.apps.common.presentation.charts.layout import ChartPaneLayout, DataTableLayout
from dao_analyzer.web.apps.common.presentation.charts.layout.figure import Figure, BarFigure, CalFigure, MultiBarFigure, TreemapFigure
from dao_analyzer.web.apps.common.business.i_metric_adapter import IMetricAdapter
from dao_analyzer.web.apps.common.business.singleton import Singleton
from dao_analyzer.web.apps.daohaus.business.metric_adapter.basic_adapter import BasicAdapter
from dao_analyzer.web.apps.daohaus.business.metric_adapter.votes_type import VotesType
from dao_analyzer.web.apps.daohaus.business.metric_adapter.proposal_outcome import ProposalOutcome
from dao_analyzer.web.apps.daohaus.business.metric_adapter.proposal_type import ProposalType 
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.metric_dao_factory as s_factory
from dao_analyzer.web.apps.daohaus.data_access.daos.platform_dao import DaohausDao
from dao_analyzer.web.apps.daohaus.resources.strings import TEXT

class DaohausService(metaclass=Singleton):
    _MEMBER: int = 0
    _VOTE: int = 1
    _PROPOSAL: int = 3
    _ORGANIZATION: int = 4
    _ASSETS: int = 5

    def __init__(self):
        # app state
        self.__orgsDAO: PlatformDAO = DaohausDao()
        self.__controllers: Dict[int, List[ChartController]] = {
            self._MEMBER: list(),
            self._VOTE: list(),
            self._PROPOSAL: list(),
            self._ORGANIZATION: list(),
            self._ASSETS: list()
        }
        self.__already_bound: bool = False
        self.__data_points = {}


    def bind_callbacks(self, app) -> None:
        if not self.__already_bound:
            self.__already_bound = True
            view_cont.bind_callbacks(
                app=app,
                section_id=TEXT['css_id_organization'],
                )
            self.__gen_sections()

            for contList in self.__controllers.values():
                for c in contList:
                    if hasattr(c, 'bind_callback'):
                        c.bind_callback(app)

    def platform(self, orglist: OrganizationList) -> Platform:
        return self.__orgsDAO.get_platform(orglist)

    def organization_list(self) -> OrganizationList:
        return self.__orgsDAO.get_organization_list()

    @property
    def are_panes(self) -> bool:
        """
        Checks if panes and their controllers are already created.
        """
        return any(self.__controllers.values())


    def get_layout(self, **kwargs) -> html.Div:
        """
        Returns the app's layout. 
        """
        if not self.__already_bound:
            self.bind_callbacks()

        org_list = self.organization_list()

        return view.generate_layout(
            organization_list=org_list,
            platform_info=self.platform(org_list),
            sections=self.__get_sections(),
            update=self.__orgsDAO.get_last_update_str(),
            platform_id=TEXT['css_id_organization'],
            datapoints=self.__get_datapoints(),
            **kwargs,
        )


    def __gen_sections(self) -> None:
        self.__get_member_charts()
        self.__get_vote_charts()
        self.__get_proposal_charts()
        self.__get_organization_charts()
        self.__get_assets_charts()

    def __get_sections(self) -> Dict[str, List[Callable]]:
        """
        Returns a dict with each section filled with a callable function which
         returns the chart layout
        """
        l_member: List[Callable] = list()
        l_vote: List[Callable] = list()
        l_proposal: List[Callable] = list()
        l_organization: List[Callable] = list()
        l_assets: List[Callable] = list()

        if not self.are_panes:
            self.__gen_sections()

        # Panes are already created.
        l_member = [c.layout.get_layout for c in self.__controllers[self._MEMBER]]
        l_vote = [c.layout.get_layout for c in self.__controllers[self._VOTE]]
        l_proposal = [c.layout.get_layout for c in self.__controllers[self._PROPOSAL]]
        l_organization = [c.layout.get_layout for c in self.__controllers[self._ORGANIZATION]]
        l_assets = [c.layout.get_layout for c in self.__controllers[self._ASSETS]]

        return {
            TEXT['title_activity']: {
                'callables': l_organization,
                'css_id': TEXT['css_id_activity'],
                'disclaimer': TEXT['disclaimer_activity']
            },
            TEXT['title_member']: {
                'callables': l_member,
                'css_id': TEXT['css_id_member'],
            },
            TEXT['title_proposal']: {
                'callables': l_proposal,
                'css_id': TEXT['css_id_proposal'],
            },
            TEXT['title_vote']: {
                'callables': l_vote,
                'css_id': TEXT['css_id_vote'],
            },
            TEXT['title_assets']: {
                'callables': l_assets,
                'css_id': TEXT['css_id_assets']
            }
        }

    def __get_datapoints(self):
        if not self.are_panes:
            self.__gen_sections()

        return self.__data_points.values()

    def __get_organization_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # active organizations
        charts.append(self.__create_chart(
            title=TEXT['title_active_organization'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_ORGANIZATION, 
            ),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION,
            css_classes=['only-on-all-orgs']
        ))
        
        charts.append(self.__create_chart(
            title=TEXT['title_organization_activity'],
            adapter=BasicAdapter(
                metric_id=s_factory.ORGANIZATION_ACTIVITY,
            ),
            figure=CalFigure(),
            cont_key=self._ORGANIZATION
        ))
        
        return charts



    def __get_member_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # new members
        charts.append(self.__create_chart(
            title=TEXT['title_new_members'],
            adapter=BasicAdapter(
                metric_id=s_factory.NEW_MEMBERS, 
            ),
            figure=BarFigure(),
            cont_key=self._MEMBER
        ))

        # rage quits
        charts.append(self.__create_chart(
            title=TEXT['title_out_members'],
            adapter=BasicAdapter(
                metric_id=s_factory.OUTGOING_MEMBERS, 
            ),
            figure=BarFigure(),
            cont_key=self._MEMBER
        ))

        # total members
        charts.append(self.__create_sum_chart(
            title=TEXT['title_total_members'],
            adapter=BasicAdapter(
                metric_id=s_factory.TOTAL_MEMBERS, 
            ),
            figure=BarFigure(),
            cont_key=self._MEMBER,
            dp_id=TEXT['dp_id_members'],
            dp_title=TEXT['dp_title_members'],
        ))

        # active members
        charts.append(self.__create_sum_chart(
            title=TEXT['title_active_members'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_MEMBERS, 
            ),
            figure=BarFigure(),
            cont_key=self._MEMBER,
            dp_id=TEXT['dp_id_active_members'],
            dp_title=TEXT['dp_title_active_members'],
        ))

        return charts


    def __get_vote_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # votes by type
        charts.append(self.__create_chart(
            title=TEXT['title_vote_type'],
            adapter=VotesType(),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._VOTE
        ))

        # votes for rate
        charts.append(self.__create_chart(
            title=TEXT['title_vote_for_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.VOTES_FOR_RATE, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()

        # votes against rate
        charts.append(self.__create_chart(
            title=TEXT['title_vote_against_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.VOTES_AGAINST_RATE, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()

        # active voters
        charts.append(self.__create_chart(
            title=TEXT['title_active_voters'],
            adapter=BasicAdapter(
                metric_id=s_factory.ACTIVE_VOTERS, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))

        # percentage of reputation holders which vote
        charts.append(self.__create_chart(
            title=TEXT['title_voters_percentage'],
            adapter=BasicAdapter(s_factory.VOTERS_PERCENTAGE),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        self.__controllers[self._VOTE][-1].layout.configuration.disable_subtitles()
        self.__controllers[self._VOTE][-1].layout.figure\
            .configuration.add_y_params(params={'suffix': '%'})

        # votes-voters rate
        charts.append(self.__create_chart(
            title=TEXT['title_votes_voters'],
            adapter=BasicAdapter(
                metric_id=s_factory.VOTES_VOTERS_RATE, 
            ),
            figure=BarFigure(),
            cont_key=self._VOTE
        ))
        return charts


    def __get_proposal_charts(self) -> List[Callable[[], html.Div]]:
        charts: List[Callable] = list()

        # new proposals
        charts.append(self.__create_sum_chart(
            title=TEXT['title_new_proposals'],
            adapter=BasicAdapter(
                metric_id=s_factory.NEW_PROPOSALS, 
            ),
            figure=BarFigure(),
            cont_key=self._PROPOSAL,
            dp_id=TEXT['dp_id_proposals'],
            dp_title=TEXT['dp_title_proposals'],
        ))

        # proposal outcomes
        charts.append(self.__create_chart(
            title=TEXT['title_proposal_outcomes'],
            adapter=ProposalOutcome(),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._PROPOSAL
        ))

        # approval proposal rate
        charts.append(self.__create_chart(
            title=TEXT['title_approval_proposal_rate'],
            adapter=BasicAdapter(
                metric_id=s_factory.APPROVAL_PROPOSAL_RATE, 
            ),
            figure=BarFigure(),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()

        # proposal types
        charts.append(self.__create_chart(
            title=TEXT['title_proposal_type'],
            adapter=ProposalType(),
            figure=MultiBarFigure(bar_type=MultiBarFigure.STACK),
            cont_key=self._PROPOSAL
        ))
        self.__controllers[self._PROPOSAL][-1].layout.configuration.disable_subtitles()
        
        return charts

    
    def __get_assets_charts(self):
        charts: List[Callable] = list()

        # assets with value
        # TODO: Show value in $ in the title
        charts.append(self.__create_sum_chart(
            title=TEXT['title_assets_value'],
            adapter=AssetsValues(),
            figure=TreemapFigure(),
            cont_key=self._ASSETS,
            dp_id=TEXT['dp_id_treasury'],
            dp_title=TEXT['dp_title_treasury'],
        ))
        self.__controllers[self._ASSETS][-1].layout.configuration.disable_subtitles()

        # assets without value (other tokens)
        charts.append(self.__create_dataTable(
            title=TEXT['title_assets_novalue'],
            adapter=AssetsTokens(),
            cont_key=self._ASSETS
        ))
 
        return charts

    def __create_chart(self, title: str, adapter: IMetricAdapter, figure: Figure
    , cont_key: int, css_classes: List[str] = [], showOnlyOnAll: bool = False) -> Callable:
        """
        Creates the chart layout and its controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure,
            css_classes=css_classes,
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        controller: ChartController = ChartController(
            css_id=css_id,
            layout=layout,
            adapter=adapter,
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout

    def __create_sum_chart(self, 
        title: str,
        adapter: IMetricAdapter,
        figure: Figure,
        cont_key: int,
        dp_id: str,
        dp_title: str,
    ):
        """
        Creates the chart layout and its summary controller, and returns a callable
        to get the html representation.
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"
        layout: ChartPaneLayout = ChartPaneLayout(
            title=title,
            css_id=css_id,
            figure=figure
        )
        layout.configuration.set_css_border(css_border=TEXT['css_pane_border'])

        self.__data_points[dp_id] = dac.DataPoint(
            id=dp_id,
            title=dp_title,
        )

        controller: ChartController = ChartSummaryController(
            css_id=css_id,
            layout=layout,
            adapter=adapter,
            dp_id=dp_id,
        )

        self.__controllers[cont_key].append(controller)

        return layout.get_layout
        

    def __create_dataTable(self, title: str, adapter: IMetricAdapter, cont_key: int) -> Callable:
        """Creates a datatable to put alongside charts

        Args:
            title (str): The title of the datatable
            adapter (IMetricAdapter): The adapter to get the data from
            cont_key (int): The key of the controller

        Returns:
            Callable: Layout html builder
        """
        css_id: str = f"{TEXT['pane_css_prefix']}{ChartPaneLayout.pane_id()}"

        layout: DataTableLayout = DataTableLayout(
            title=title,
            css_id=css_id
        )

        controller: DataTableController = DataTableController(
            table_id=layout.table_id,
            layout=layout,
            adapter=adapter
        )

        self.__controllers[cont_key].append(controller)
        return layout.get_layout



from .bar_figure import BarFigure
from .calplot import CalFigure
from .double_scatter_figure import DoubleScatterFigure
from .figure import Figure
from .multi_bar_figure import MultiBarFigure
from .treemap_figure import TreemapFigure

__all__ = [
    'Figure',
    'BarFigure',
    'CalFigure',
    'DoubleScatterFigure',
    'MultiBarFigure',
    'TreemapFigure',
]


from typing import Dict

from dao_analyzer.web.apps.common.business.transfers import TabularData, OrganizationList
from dao_analyzer.web.apps.common.business.i_metric_adapter import IMetricAdapter
import dao_analyzer.web.apps.daohaus.data_access.daos.metric.metric_dao_factory as s_factory

class AssetsTokens(IMetricAdapter):
    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param
        """
        dao: TabularData = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=s_factory.ASSETS_TOKENS
        )

        return dao.get_metric()